import xml.etree.ElementTree as ET
import uuid
from xml.dom import minidom

def generate_coords_xml(coordinates, group_name=None):
    """
    Generates XML simulating a Leica .rgn file

    Parameters
    ----------
    coordinates : np.ndarray
        Numpy array containing coordinates, must be of shape (len, 2)
    group_name : str | None, default=None
        Group name to save the regions under

    Returns
    -------
    pretty_xml : str
        XML string containing coordinates
    """

    # Create root element
    root = ET.Element("StageOverviewRegions")
    
    # Create Regions element
    regions = ET.SubElement(root, "Regions")
    shape_list = ET.SubElement(regions, "ShapeList")
    items = ET.SubElement(shape_list, "Items")
    
    # Store identifiers to use later in StackList
    identifiers = []
    
    # Check if points should be grouped
    if group_name:
        # Create a compound shape to hold the points
        compound_item = ET.SubElement(items, "Item0")
        
        # Set group attributes
        ET.SubElement(compound_item, "Name").text = group_name
        compound_id = str(uuid.uuid4())
        ET.SubElement(compound_item, "Identifier").text = compound_id
        ET.SubElement(compound_item, "Type").text = "CompoundShape"
        ET.SubElement(compound_item, "Fill").text = "R:1 ,G: 0,B: 0,A: 0"
        ET.SubElement(compound_item, "Font")
        
        # Create empty verticies
        verticies = ET.SubElement(compound_item, "Verticies")
        ET.SubElement(verticies, "Items")
        
        # Create empty decorator colors
        decorator_colors = ET.SubElement(compound_item, "DecoratorColors")
        ET.SubElement(decorator_colors, "Items")
        
        # Create empty extended properties
        extended_props = ET.SubElement(compound_item, "ExtendedProperties")
        ET.SubElement(extended_props, "Items")
        
        # Create children container
        children = ET.SubElement(compound_item, "Children")
        children_items = ET.SubElement(children, "Items")
        
        # Points will be added to the children_items
        points_container = children_items
    else:
        # Points will be added directly to the main items
        points_container = items
    
    # Add point items
    for i, (x, y) in enumerate(coordinates):
        # Create item element with proper name (Item0, Item1, etc.)
        item = ET.SubElement(points_container, f"Item{i}")
        
        # Generate a unique identifier (UUID)
        identifier = str(uuid.uuid4())
        identifiers.append(identifier)
        
        # Create subelements for the item
        ET.SubElement(item, "Number").text = str((i + 1) * 2 + 7)  # Just an example increment
        
        # Add Name element if using CompoundShape
        if group_name:
            ET.SubElement(item, "Name").text = f"P{i+1}"
            
        ET.SubElement(item, "Tag").text = f"P {i + 1}"
        ET.SubElement(item, "Identifier").text = identifier
        ET.SubElement(item, "Type").text = "Point"
        ET.SubElement(item, "Fill").text = "R:1 ,G: 0,B: 0,A: 0"
        ET.SubElement(item, "Font")
        
        # Create verticies element
        verticies = ET.SubElement(item, "Verticies")
        verticies_items = ET.SubElement(verticies, "Items")
        vertex_item = ET.SubElement(verticies_items, "Item0")
        
        # Add X and Y coordinates
        ET.SubElement(vertex_item, "X").text = f"{x:.11f}"
        ET.SubElement(vertex_item, "Y").text = f"{y:.11f}"
        
        # Add empty elements
        decorator_colors = ET.SubElement(item, "DecoratorColors")
        ET.SubElement(decorator_colors, "Items")
        extended_props = ET.SubElement(item, "ExtendedProperties")
        ET.SubElement(extended_props, "Items")
    
    # Add attributes to ShapeList
    ET.SubElement(shape_list, "FillMaskMode").text = "None"
    ET.SubElement(shape_list, "VertexUnitMode").text = "Pixels"
    
    # Create StackList
    stack_list = ET.SubElement(root, "StackList")
    
    # Add entries to StackList
    for identifier in identifiers:
        entry = ET.SubElement(stack_list, "Entry")
        entry.set("Identifier", identifier)
        entry.set("Begin", "0.0000000000")
        entry.set("End", "0.0000000000")
        entry.set("SectionCount", "0")
        entry.set("ReferenceX", "0.0000000000")
        entry.set("ReferenceY", "0.0000000000")
        entry.set("FocusStabilizerOffset", "0.0000000000")
        entry.set("FocusStabilizerOffsetFixed", "false")
        entry.set("StackValid", "false")
        entry.set("Marked", "false")
    
    # Convert the ElementTree to a string with pretty formatting
    rough_string = ET.tostring(root, 'utf-8')
    reparsed = minidom.parseString(rough_string)
    pretty_xml = reparsed.toprettyxml(indent="  ")
    
    return pretty_xml


# Default imports
from copy import deepcopy

# External imports 
import numpy as np
from skimage.measure import regionprops_table
import pandas as pd

# Package imports
from leica_roi_finder.core.LIF_metadata import read_lif_metadata
from leica_roi_finder.core.coords_to_xml import generate_coords_xml
from leica_roi_finder.core._defaults import(
    DEFAULT_CELLPROB_THRESHOLD,
    DEFAULT_DIAMETER,
    DEFAULT_FLOW_THRESHOLD,
    DEFAULT_MIN_INTENSITY,
    DEFAULT_MAX_INTENSITY,
    DEFAULT_MIN_SIZE,
    DEFAULT_MAX_SIZE,
    DEFAULT_MIN_CIRCULARITY,
    DEFAULT_MAX_CIRCULARITY
)

class ROI_finder():
    """"
    Automatic ROI finder using cellpose and other parameters

    Attributes
    ----------
    diameter : int
        Cellpose diameter
    flow_threshold : float
        Cellpose flow threshold
    cellprob_threshold : float
        Cellpose cell probability threshold
    min_intensity : int
        Minimal average intensity a cell should have
    max_intensity : int
        Maximum average intensity a cell should have
    min_size : int
        The minimal amount of pixels a cell should have
    max_size : int
        The maximum amount of pixels a cell should have
    min_circularity : float
        The minimal circularity a cell should have, circularity ranges from 0 to 1 (perfect circle)
    max_circularity : float
        The maximum circularity a cell should have, circularity ranges from 0 to 1 (perfect circle)
    segmented_mask : ndarray
        Original mask segmented by cellpose
    mask : ndarray
        Mask segmented by cellpose, but removed cells that did not fullfil other criteria (e.g. min/max intensity)
    img : ndarray
        Image extracted from .lif file
    metadata : dict
        Relevant metadata extracted from .lif file
    coords : ndarray
        Numpy array containing coordinates of ROIs, array is of shape (len, 2)

    Notes
    -----
    Not all attributes will be available/calculated until the class is properly initialized using the "run()" method
    Parameters such as diameter can be set during construction, or altered later.
    Once parameters have changed, you can use the "run()" method to calculate the new ROIs
    """

    def __init__(
            self,
            filepath,
            diameter = DEFAULT_DIAMETER,
            flow_threshold = DEFAULT_FLOW_THRESHOLD,
            cellprob_threshold = DEFAULT_CELLPROB_THRESHOLD,
            min_intensity = DEFAULT_MIN_INTENSITY,
            max_intensity = DEFAULT_MAX_INTENSITY,
            min_size = DEFAULT_MIN_SIZE,
            max_size = DEFAULT_MAX_SIZE,
            min_circularity = DEFAULT_MIN_CIRCULARITY,
            max_circularity = DEFAULT_MAX_CIRCULARITY
        ):
        """"
        Parameters
        ----------
        filepath : str
            File path for *.lif file
        diameter : int
            Cellpose diameter
        flow_threshold : float
            Cellpose flow threshold
        cellprob_threshold : float
            Cellpose cell probability threshold
        min_intensity : int
            Minimal average intensity a cell should have
        max_intensity : int
            Maximum average intensity a cell should have
        min_size : int
            The minimal amount of pixels a cell should have
        max_size : int
            The maximum amount of pixels a cell should have
        min_circularity : float
            The minimal circularity a cell should have, circularity ranges from 0 to 1 (perfect circle)
        max_circularity : float
            The maximum circularity a cell should have, circularity ranges from 0 to 1 (perfect circle)
        """

        self._filepath = filepath

        # Share parameters across class
        self.diameter = diameter
        self.flow_threshold = flow_threshold
        self.cellprob_threshold = cellprob_threshold
        self.min_intensity = min_intensity
        self.max_intensity = max_intensity
        self.min_size = min_size
        self.max_size = max_size
        self.min_circularity = min_circularity
        self.max_circularity = max_circularity

        # Store parameters of previous run
        self._previous_diameter = None
        self._previous_flow_threshold = None
        self._previous_cellprob_threshold = None
        self._previous_min_intensity = None
        self._previous_max_intensity = None
        self._previous_min_size = None
        self._previous_max_size = None
        self._previous_min_circularity = None
        self._previous_max_circularity = None

        # Read lif
        self._read_lif()

        # Variables
        self.model=None
        self.segmented_mask = np.zeros_like(self.img)
        self.mask = np.zeros_like(self.img)
        self.coords = []

    def run(self):
        """
        Run cellpose segmentation and other criteria

        Notes
        -----
        If parameters for cellpose segmentation have not changed, the segmentation will not be run again to save time.
        """

        run_select_roi=False

        # Check if cellpose parameters have changed - if they have run segmentation again
        if (
            self._previous_diameter != self.diameter or
            self._previous_flow_threshold != self.flow_threshold or
            self._previous_cellprob_threshold != self.cellprob_threshold
        ):
            # Run cellpose
            self._cellpose()
            run_select_roi=True # Now also rerun ROI selection

        # Sort cells base on other criteria
        # If nothing has changed, do not run ROI selection
        if (
            self._previous_min_intensity != self.min_intensity or
            self._previous_max_intensity != self.max_intensity or
            self._previous_min_size != self.min_size or
            self._previous_max_size != self.max_size or
            self._previous_min_circularity != self.min_circularity or
            self._previous_max_circularity != self.max_circularity
        ) or (
            run_select_roi
        ):
            self._select_roi()

        # Store parameters
        self._previous_diameter = deepcopy(self.diameter)
        self._previous_flow_threshold = deepcopy(self.flow_threshold)
        self._previous_cellprob_threshold = deepcopy(self.cellprob_threshold)
        self._previous_min_intensity = deepcopy(self.min_intensity)
        self._previous_max_intensity = deepcopy(self.max_intensity)
        self._previous_min_size = deepcopy(self.min_size)
        self._previous_max_size = deepcopy(self.max_size)
        self._previous_min_size = deepcopy(self.min_circularity)
        self._previous_max_size = deepcopy(self.max_circularity)

    def _read_lif(self):
        """
        Read lif metadata and assign values to attributes
        """

        # Read metadata
        metadata = read_lif_metadata(self._filepath)

        if metadata["MicroscopeModel"] != "DMI8-CS":
            raise UserWarning(f"This script was created for the Leica Stellaris 8 microscope, and has only been tested using the Leica Stellaris 8. Using this script on other machines might lead to unexpected outputs due to metadata differences")

        self.dtype =  getattr(np, f'uint{metadata["BitSize"]}')

        # Load memmap
        lif_img = np.memmap(
            self._filepath,
            offset=metadata["Offset"],
            mode='r',
            shape=(metadata["YDim"], metadata["XDim"]),
            dtype=self.dtype
        )

        # Flip image if necessary
        self.flipy = metadata["FlipY"]
        if self.flipy:
            lif_img = np.flipud(lif_img)
        self.flipx = metadata["FlipX"]
        if self.flipx:
            lif_img = np.fliplr(lif_img)

        self.img = lif_img
        self.metadata = metadata
        self._posy = metadata["PosY"]
        self._posx = metadata["PosX"]

    def _cellpose(self):
        """
        Perform segmentation using cellpose
        """

        # Load model if not done yet
        if self.model == None:
            # Lazy imports
            print("Loading cellpose...")
            from cellpose import models
            from torch.cuda import is_available
            if is_available():
                gpu=True
                print("Using GPU...")
            else:
                gpu=False
                print("Using CPU...")
            print("Loading model...")
            self.model = models.Cellpose(gpu=gpu, model_type='cyto3')

        # Run model
        self.mask, flow, styles, diams = self.model.eval(self.img, diameter=self.diameter, flow_threshold=self.flow_threshold, cellprob_threshold=self.cellprob_threshold)
        self.segmented_mask = deepcopy(self.mask)

    def _select_roi(self):
        """
        Select regions of interest based on intensity, size and roundness
        """
        self.mask = self.segmented_mask.copy()

        # Calculate builtin regionprops properties
        props = regionprops_table(
            self.segmented_mask, 
            intensity_image=self.img,
            properties=['label', 'area', 'perimeter', 'mean_intensity']
        )
        df = pd.DataFrame(props)

        # Calculate circularity
        df['circularity'] = 4 * np.pi * df['area'] / (df['perimeter'] ** 2)
        df['circularity'] = df['circularity'].fillna(0)

        # Filter using conditions
        keep = df[
            (df['mean_intensity'] >= self.min_intensity) &
            (df['mean_intensity'] <= self.max_intensity) &
            (df['area'] >= self.min_size) &
            (df['area'] <= self.max_size) &
            (df['circularity'] >= self.min_circularity) &
            (df['circularity'] <= self.max_circularity)
        ]['label'].values

        # Remove masks
        self.mask[~np.isin(self.segmented_mask, keep)] = 0

    def _find_center(self):
        """
        Find centers of selected ROIs
        """
        temp_mask = self.mask.copy()
        if self.flipy:
            temp_mask = np.flipud(temp_mask)
        if self.flipx:
            temp_mask = np.fliplr(temp_mask)

        labels, counts = np.unique(temp_mask, return_counts=True)
        if labels[0] == 0:
            labels = labels[1:]
            counts = counts[1:]
        
        centers = []
        y_coords, x_coords = np.indices(temp_mask.shape)
        
        for label in labels:
            mask = (temp_mask == label)
            center_y = np.mean(y_coords[mask])
            center_x = np.mean(x_coords[mask])
            centers.append([int(center_x), int(center_y)])
        
        return np.array(centers)
    
    def export_to_rgn(self, outputpath, groupname=None):
        """
        Convert ROIs to Leica coordinates and save to .rgn file

        Parameters
        ----------
        outputpath : str
            Filepath where to save the region file. Should have suffix ".rgn"
        groupname : str, optional
            Group name inside the xml
        """
        xres = self.metadata["XRes"]  # Meters
        yres = self.metadata["YRes"]  # Meters
        
        # Find center of ROIs
        self.coords = self._find_center()

        # Calculate stage coordinates relative to 0, 0 in image
        leica_coords = []
        for pos in self.coords:
            leica_xcoord = pos[0]*xres
            leica_ycoord = pos[1]*yres
            leica_coords.append([leica_xcoord, leica_ycoord])
        leica_coords = np.array(leica_coords)

        # Recalculate coordinates relative to middle of image
        self._posx = self._posx - (self.img.shape[1]/2)*xres
        self._posy = self._posy - (self.img.shape[0]/2)*yres
        
        # Calculate absolute stage position and add to coordinates
        leica_coords[:, 0] += self._posx
        leica_coords[:, 1] += self._posy

        xml = generate_coords_xml(leica_coords, group_name=groupname)

        with open(outputpath, 'w') as f:
            f.write(xml)


# Default imports
import sys
import traceback
from pathlib import Path

# External imports
import numpy as np

# Package imports
from leica_roi_finder.core.roi_finder_class import ROI_finder
from leica_roi_finder.core._defaults import(
    DEFAULT_CELLPROB_THRESHOLD,
    DEFAULT_DIAMETER,
    DEFAULT_FLOW_THRESHOLD,
    DEFAULT_MIN_INTENSITY,
    DEFAULT_MAX_INTENSITY,
    DEFAULT_MIN_SIZE,
    DEFAULT_MAX_SIZE,
    DEFAULT_MIN_CIRCULARITY,
    DEFAULT_MAX_CIRCULARITY
)
import leica_roi_finder

# PyQt6 imports
from PyQt6.QtWidgets import (
    QApplication, QMainWindow, QWidget,
    QPushButton, QLabel, QLineEdit, QGroupBox,
    QFileDialog, QGridLayout, QMessageBox,
    QHBoxLayout, QVBoxLayout, QLabel,
    QSlider, QPushButton, QGroupBox
    )

from PyQt6.QtCore import pyqtSlot, Qt
from PyQt6.QtGui import QAction, QPalette, QColor

# Other QT imports
import pyqtgraph as pg
from pyqtgraph.widgets.HistogramLUTWidget import HistogramLUTWidget
from superqt import QRangeSlider

class main_window(QMainWindow):
    """
    Main window for Leica ROI selector
    """
    def __init__(self):
        super().__init__()

        self.img = np.zeros((600, 600))
        self.roi_mask = np.zeros((600, 600))

        self.window_title = f"Leica ROI finder - Version: {leica_roi_finder.__version__}"
        self.setWindowTitle(self.window_title)

        # Variables
        self.bit_depth = 16
        self.loaded_image = False

        self.init_ui()

    def init_ui(self):
        """
        Initialize main UI components (menubar, sidebar, image display)
        """
        # Create menu bar
        self.create_menubar()

        # Create main widget and layout
        main_widget = QWidget()
        main_layout = QHBoxLayout(main_widget)
        self.setCentralWidget(main_widget)
        
        # Create sidebar widget
        self.sidebar = self.create_sidebar()
        main_layout.addWidget(self.sidebar)
        
        # Create image display widget
        self.image_widget = self.create_image_display()
        main_layout.addWidget(self.image_widget, 1)
        
        # Initialize
        self.update_display()

    def create_menubar(self):
        """
        Initialize menubar
        """
        # Create menu bar
        menubar = self.menuBar()
        
        # Create file menu
        file_menu = menubar.addMenu("File")
        
        # Create open action
        open_action = QAction("Open file", self)
        open_action.setShortcut("Ctrl+O")
        open_action.triggered.connect(self.open_image)
        file_menu.addAction(open_action)
        
        # Create exit action
        exit_action = QAction("Exit", self)
        exit_action.setShortcut("Ctrl+Q")
        exit_action.triggered.connect(self.close)
        file_menu.addAction(exit_action)

    def create_sidebar(self):
        """
        Initialize sidebar
        """
        sidebar_widget = QWidget()
        sidebar_layout = QVBoxLayout(sidebar_widget)
        sidebar_layout.setContentsMargins(10, 10, 10, 10)
        
        # Create groups for controls
        image_controls = QGroupBox("Image Controls")
        image_layout = QVBoxLayout(image_controls)

        # Image controls
        image_label = QLabel("Min/Max:")
        self.image_slider = QRangeSlider(Qt.Orientation.Horizontal)
        self.image_slider.setMinimum(int(np.min(self.img)))
        self.image_slider.setMaximum(int(np.max(self.img)))
        self.image_slider.setValue((int(np.min(self.img)), int(np.max(self.img))))
        self.image_slider.valueChanged.connect(self.update_levels)
        image_layout.addWidget(image_label)
        image_layout.addWidget(self.image_slider)

        opacity_label = QLabel("Mask Opacity:")
        self.opacity_slider = QSlider(Qt.Orientation.Horizontal)
        self.opacity_slider.setRange(0, 100)
        self.opacity_slider.setValue(50)
        self.opacity_slider.valueChanged.connect(self.update_mask_opacity)
        image_layout.addWidget(opacity_label)
        image_layout.addWidget(self.opacity_slider)

        # Cellpose settings group
        cellpose_controls = QGroupBox("Cellpose settings:")
        cellpose_layout = QGridLayout(cellpose_controls)

        # Diameter input
        diameter_label = QLabel("Diameter:")
        self.diameter_input = QLineEdit()
        self.diameter_input.setText(str(DEFAULT_DIAMETER))
        cellpose_layout.addWidget(diameter_label, 0, 0)
        cellpose_layout.addWidget(self.diameter_input, 0, 1)

        # Flow threshold input
        flow_label = QLabel("Flow threshold:")
        flow_label.setToolTip("Set higher to get more cells, ranges from 0.0 to 3.0")
        self.flow_input = QLineEdit()
        self.flow_input.setText(str(DEFAULT_FLOW_THRESHOLD))
        cellpose_layout.addWidget(flow_label, 1, 0)
        cellpose_layout.addWidget(self.flow_input, 1, 1)

        # Cellprob threshold input
        cellprob_label = QLabel("Cellprob threshold:")
        cellprob_label.setToolTip("Set lower to include more pixels, ranges from -6.0 to 6.0")
        self.cellprob_input = QLineEdit()
        self.cellprob_input.setText(str(DEFAULT_CELLPROB_THRESHOLD))
        cellpose_layout.addWidget(cellprob_label, 2, 0)
        cellpose_layout.addWidget(self.cellprob_input, 2, 1)

        """Intensity settings group"""
        intensity_controls = QGroupBox("Intensity")
        intensity_layout = QGridLayout(intensity_controls)

        # Min intensity input
        min_intensity_label = QLabel("Min intensity:")
        self.min_intensity_input = QLineEdit()
        self.min_intensity_input.setText(str(DEFAULT_MIN_INTENSITY))
        intensity_layout.addWidget(min_intensity_label, 0, 0)
        intensity_layout.addWidget(self.min_intensity_input, 0, 1)

        # Min intensity input
        max_intensity_label = QLabel("Max intensity:")
        self.max_intensity_input = QLineEdit()
        self.max_intensity_input.setText(str(DEFAULT_MAX_INTENSITY))
        intensity_layout.addWidget(max_intensity_label, 1, 0)
        intensity_layout.addWidget(self.max_intensity_input, 1, 1)

        """Size settings group"""
        size_controls = QGroupBox("Size (pixels)")
        size_layout = QGridLayout(size_controls)

        # Min size input
        min_size_label = QLabel("Min size:")
        self.min_size_input = QLineEdit()
        self.min_size_input.setText(str(DEFAULT_MIN_SIZE))
        size_layout.addWidget(min_size_label, 0, 0)
        size_layout.addWidget(self.min_size_input, 0, 1)

        # Max size input
        max_size_label = QLabel("Max size:")
        self.max_size_input = QLineEdit()
        self.max_size_input.setText(str(DEFAULT_MAX_SIZE))
        size_layout.addWidget(max_size_label, 1, 0)
        size_layout.addWidget(self.max_size_input, 1, 1)

        """Circularity settings group"""
        circularity_controls = QGroupBox("Circularity (0-1)")
        circularity_layout = QGridLayout(circularity_controls)

        # Min circularity input
        min_circularity_label = QLabel("Min circularity:")
        self.min_circularity_input = QLineEdit()
        self.min_circularity_input.setText(str(DEFAULT_MIN_CIRCULARITY))
        circularity_layout.addWidget(min_circularity_label, 0, 0)
        circularity_layout.addWidget(self.min_circularity_input, 0, 1)

        # Max circularity input
        max_circularity_label = QLabel("Max circularity:")
        self.max_circularity_input = QLineEdit()
        self.max_circularity_input.setText(str(DEFAULT_MAX_CIRCULARITY))
        circularity_layout.addWidget(max_circularity_label, 1, 0)
        circularity_layout.addWidget(self.max_circularity_input, 1, 1)

        # Detected ROIs label
        self.detected_rois_label = QLabel()
        self.detected_rois_label.setText(f'<b><span style="font-size:15px;">Detected ROIs: 0</span></b>')

        # Button to run the segmentation/roi selection
        run_button = QPushButton("Run ROI finder")
        run_button.clicked.connect(self.run)

        # Export button
        export_button = QPushButton("Export to .rgn")
        export_button.clicked.connect(self.export)
        
        # Add everything to sidebar
        sidebar_layout.addWidget(image_controls)
        sidebar_layout.addWidget(cellpose_controls)
        sidebar_layout.addWidget(intensity_controls)
        sidebar_layout.addWidget(size_controls)
        sidebar_layout.addWidget(circularity_controls)
        sidebar_layout.addWidget(self.detected_rois_label)
        sidebar_layout.addStretch(1)  # Push everything up
        sidebar_layout.addWidget(run_button)
        sidebar_layout.addWidget(export_button)

        # Collect all QLineEdits and bind the "run" command
        qlineedits = [
            self.diameter_input,
            self.flow_input,
            self.cellprob_input,
            self.min_intensity_input,
            self.max_intensity_input,
            self.min_size_input,
            self.max_size_input,
            self.min_circularity_input,
            self.max_circularity_input,
        ]

        for lineedit in qlineedits:
            lineedit.returnPressed.connect(self.run)

        
        return sidebar_widget
    
    def create_image_display(self):
        """
        Initialize image display
        """
        # Create a GraphicsLayoutWidget
        graphics_layout = pg.GraphicsLayoutWidget()
        
        # Add a ViewBox
        self.view_box = graphics_layout.addViewBox()
        self.view_box.setAspectLocked(True)
        
        # Create ImageItem for the base image
        self.image_item = pg.ImageItem()
        self.view_box.addItem(self.image_item)
        
        # Create ImageItem for the mask overlay
        self.mask_item = pg.ImageItem()
        self.view_box.addItem(self.mask_item)
        
        # Create and link a HistogramLUTWidget
        self.histogram_widget = HistogramLUTWidget()
        self.histogram_widget.setImageItem(self.image_item)
        self.histogram_widget.gradient.loadPreset("grey")

        self.histogram_widget.sigLevelChangeFinished.connect(self.update_levels_from_histogram)

        container = QWidget()
        layout = QHBoxLayout(container)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(graphics_layout, 14)
        layout.addWidget(self.histogram_widget, 1)

        return container
    
    def update_levels_from_histogram(self):
        """
        Update image display based on histogram slider
        """
        levels = self.histogram_widget.getLevels()
        if levels:
            self.image_item.setLevels(levels)
            self.image_slider.setValue(tuple(map(int, levels)))  # Sync QRangeSlider

    def update_levels(self, values):
        """
        Update image display based on sidebar Min/Max slider
        """
        min_val, max_val = values
        self.image_item.setLevels([min_val, max_val])
        self.histogram_widget.setLevels(min_val, max_val)  # Sync histogram

    def update_mask_opacity(self, value):
        """
        Update segmentation mask opacity
        """
        opacity = value / 100.0
        self.mask_item.setOpacity(opacity)

    def update_display(self):
        """
        Display mask and image
        """
        self.image_item.setImage(self.img.T)
        self.mask_item.setImage(self.roi_mask.T)
        self.histogram_widget.setImageItem(self.image_item)  # Ensure link
        self.histogram_widget.setLevels(np.min(self.img), np.max(self.img))
    
    def open_image(self):
        """
        Open image file and initialize ROI_finder class
        """
        file_path, _ = QFileDialog.getOpenFileName(
            self,
            "Open LIF file",
            "",
            "Image Files (*.lif)"
        )
        
        if file_path:
            try:
                self.ROI_finder = ROI_finder(filepath=file_path)
                self.img = self.ROI_finder.img
                self.roi_mask = self.ROI_finder.mask
                
                self.bit_depth = int(self.ROI_finder.metadata["BitSize"])

                # Mask colormap
                rng = np.random.default_rng()
                lut = rng.integers(0, 256, size=(2**self.bit_depth, 3), dtype=np.uint8)
                
                alpha = np.full((2**self.bit_depth, 1), 255, dtype=np.uint8)
                alpha[0] = 0  # Make value 0 transparent
                
                rgba_lut = np.hstack((lut, alpha))
                self.mask_item.setLookupTable(rgba_lut)

                self.image_slider.setMinimum(int(np.min(self.img)))
                self.image_slider.setMaximum(int(np.max(self.img)))
                self.image_slider.setValue((int(np.min(self.img)), int(np.max(self.img))))

                self.max_intensity_input.setText(str(2**self.bit_depth))

                self.max_size_input.setText(str(self.img.shape[0]*self.img.shape[1]))

                self.update_display()
                file_path = Path(file_path)
                self.setWindowTitle(f"{self.window_title} - {file_path.name}")
                self.loaded_image = True

            except Exception as e:
                QMessageBox.critical(self, "Error", f"Failed to load image: {str(e)}")
                traceback.print_exc()

    def run(self):
        """
        Pass parameters to ROI_finder class and run segmentation/criteria
        """
        # Check if ROI finder has been initialized
        if not self.loaded_image:
            return

        try:
            # Collect values and alter ROI finder object
            self.ROI_finder.diameter = int(self.diameter_input.text())
            self.ROI_finder.flow_threshold = float(self.flow_input.text())
            self.ROI_finder.cellprob_threshold = float(self.cellprob_input.text())
            self.ROI_finder.min_intensity = int(self.min_intensity_input.text())
            self.ROI_finder.max_intensity = int(self.max_intensity_input.text())
            self.ROI_finder.min_size = int(self.min_size_input.text())
            self.ROI_finder.max_size = int(self.max_size_input.text())
            self.ROI_finder.min_circularity = float(self.min_circularity_input.text())
            self.ROI_finder.max_circularity = float(self.max_circularity_input.text())
        except Exception as error:
            msg_box = QMessageBox(self)
            msg_box.setWindowTitle("Error")
            msg_box.setText(f"Could not convert all inputs to the correct datatype:\n{error}")
            msg_box.setIcon(QMessageBox.Icon.Critical)
            msg_box.show()
            return

        # Run detection
        self.ROI_finder.run()

        # Update image
        self.roi_mask = self.ROI_finder.mask
        self.mask_item.setImage(self.roi_mask.T)
        self.update_mask_opacity(self.opacity_slider.value())
        self.detected_rois_label.setText(f'<b><span style="font-size:15px;">Detected ROIs: {len(np.unique(self.roi_mask))-1}</span></b>')

    def export(self):
        """
        Export selected regions to Leica .rgn format
        """
        # Check if there are coords
        if not self.loaded_image:
            return
        if len(np.unique(self.ROI_finder.mask))-1 < 1:
            msg_box = QMessageBox(self)
            msg_box.setWindowTitle("Error")
            msg_box.setText(f"No regions found")
            msg_box.setIcon(QMessageBox.Icon.Critical)
            msg_box.show()
            return

        file_filter = "Region Files (*.rgn)"
        filename, _ = QFileDialog.getSaveFileName(
            self,
            "Save File As",
            "custom_roi.rgn",
            file_filter
        )
        if filename:
            try:
                filename = Path(filename)
                self.ROI_finder.export_to_rgn(filename, filename.stem)
                msg_box = QMessageBox(self)
                msg_box.setWindowTitle("Saved regions")
                msg_box.setText(f"Regions succesfully saved at:\n{filename}")
                msg_box.setIcon(QMessageBox.Icon.Information)
                msg_box.show()
            except Exception as error:
                traceback.print_exc()
                msg_box = QMessageBox(self)
                msg_box.setWindowTitle("Error")
                msg_box.setText(f"Could not save regions:\n{error}")
                msg_box.setIcon(QMessageBox.Icon.Critical)
                msg_box.show()

def roi_finder_gui():
    """
    Initialize ROI_finder gui and set styles
    """

    app = QApplication(sys.argv)
    app.setStyle('Fusion')

    # Set up dark palette
    dark_palette = QPalette()
    dark_palette.setColor(QPalette.ColorRole.Window, QColor(53, 53, 53))
    dark_palette.setColor(QPalette.ColorRole.WindowText, QColor(255, 255, 255))
    dark_palette.setColor(QPalette.ColorRole.Base, QColor(25, 25, 25))
    dark_palette.setColor(QPalette.ColorRole.AlternateBase, QColor(53, 53, 53))
    dark_palette.setColor(QPalette.ColorRole.ToolTipBase, QColor(53, 53, 53))
    dark_palette.setColor(QPalette.ColorRole.ToolTipText, QColor(255, 255, 255))
    dark_palette.setColor(QPalette.ColorRole.Text, QColor(255, 255, 255))
    dark_palette.setColor(QPalette.ColorRole.Button, QColor(53, 53, 53))
    dark_palette.setColor(QPalette.ColorRole.ButtonText, QColor(255, 255, 255))
    dark_palette.setColor(QPalette.ColorRole.BrightText, QColor(255, 0, 0))
    dark_palette.setColor(QPalette.ColorRole.Link, QColor(42, 130, 218))
    dark_palette.setColor(QPalette.ColorRole.Highlight, QColor(42, 130, 218))
    dark_palette.setColor(QPalette.ColorRole.HighlightedText, QColor(0, 0, 0))

    app.setPalette(dark_palette)

    window = main_window()
    window.show()
    sys.exit(app.exec())

if __name__ == "__main__":
    roi_finder_gui()


DEFAULT_DIAMETER = 30
DEFAULT_FLOW_THRESHOLD = 0.4
DEFAULT_CELLPROB_THRESHOLD = 0
DEFAULT_MIN_INTENSITY = 0
DEFAULT_MAX_INTENSITY = 2**8    # Max value in 8-bit image
DEFAULT_MIN_SIZE = 0
DEFAULT_MAX_SIZE = 2048*2048
DEFAULT_MIN_CIRCULARITY = 0
DEFAULT_MAX_CIRCULARITY = 1


import struct
from lxml import etree

def read_lif_metadata(file_path):
    """
    Read metadata from a Leica LIF file

    Parameters
    ----------
    file_path : str
        File path of *.lif file

    Returns
    -------
    metadata : dict
        Dictionary containing relevant metadata for ROI finder

    Notes
    -----
    This function returns the following metadata that is required for the Leica ROI finder:
    - FlipX
    - FlipY
    - SwampXY
    - BitSize
    - MicroscopeModel
    - Offset
    - XDim
    - YDim
    - XRes
    - YRes
    - PosX
    - PosY

    This function has been tested and work for the Leica Stellaris 8
    Certain values might not have the same name in other microscope metadata
    """

    with open(file_path, 'rb') as f:
        # Basic LIF validation
        testvalue = struct.unpack('i', f.read(4))[0]
        if testvalue != 112:
            raise ValueError(f'Error Opening LIF-File: {file_path}')
        _ = struct.unpack('i', f.read(4))[0]
        testvalue = struct.unpack('B', f.read(1))[0]
        if testvalue != 42:
            raise ValueError(f'Error Opening LIF-File: {file_path}')
        testvalue = struct.unpack('i', f.read(4))[0]
        
        # Retrieve XML
        XMLObjDescriptionUTF16 = f.read(testvalue * 2)
        XMLObjDescription = XMLObjDescriptionUTF16.decode('utf-16')

        # Retrieve offset for memory map
        while True:
            data = f.read(4)
            testvalue = struct.unpack('i', data)[0]
            _ = struct.unpack('i', f.read(4))[0]
            testvalue = struct.unpack('B', f.read(1))[0]
            MemorySize = struct.unpack('q', f.read(8))[0]
            testvalue = struct.unpack('B', f.read(1))[0]
            if testvalue != 42:
                break
            testvalue = struct.unpack('i', f.read(4))[0]
            BlockIDLength = testvalue
            BlockIDData = f.read(BlockIDLength * 2)
            position = f.tell()

    root = etree.fromstring(XMLObjDescription)

    # Retrieve confocalsettings
    atl_element = root.find('.//ATLConfocalSettingDefinition')

    # Retrieve all relevant information
    atl_attributes = atl_element.attrib
    metadata = {
        "FlipX" : bool(int(atl_attributes["FlipX"])),
        "FlipY" : bool(int(atl_attributes["FlipY"])),
        "SwampXY" : bool(int(atl_attributes["SwapXY"])),
        "BitSize" : int(atl_attributes["BitSize"]),
        "MicroscopeModel" : str(atl_attributes["MicroscopeModel"]),
        "Offset" : position
    }

    # Retrieve dimensions
    atl_elements = root.findall('.//DimensionDescription')
    atl_attributes_list = [element.attrib for element in atl_elements]

    metadata['XDim'] = int(atl_attributes_list[0]['NumberOfElements'])
    metadata['YDim'] = int(atl_attributes_list[1]['NumberOfElements'])

    # Compute pixel size
    metadata['XRes'] = float(atl_attributes_list[0]['Length'])/(metadata['XDim']-1)
    metadata['YRes'] = float(atl_attributes_list[1]['Length'])/(metadata['YDim']-1)

    # Retrieve tilescan offset
    tile = root.find('.//Tile').attrib

    metadata["PosX"] = float(tile["PosX"])
    metadata["PosY"] = float(tile["PosY"])

    return metadata


import argparse
import os

# Package imports
from leica_roi_finder.gui.gui import roi_finder_gui
import leica_roi_finder

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--version', action='version', version=f'Leica ROI finder - version: {leica_roi_finder.__version__}')
    parser.add_argument("--shortcut", action='store_true')

    args = parser.parse_args()

    if args.shortcut:
        try:
            script_path = os.path.abspath(__file__)
            from pyshortcuts import make_shortcut
            
            make_shortcut(
                script_path, 
                name="Leica ROI finder",
                desktop=True,
                startmenu=True
            )
            print("Succesfully created desktop shortcut")
        except Exception as error:
            print(f"Failed to create shortcut:\n{error}")
    else:
        roi_finder_gui()

if __name__=="__main__":
    main()